/*

This program and source is licensed under the 

	GNU GENERAL PUBLIC LICENSE 
	
as in the included LICENSE.TXT or located at
http://www.gnu.org/licenses/gpl.txt

*/

#ifndef __ECU_H
#define __ECU_H

#include <windows.h>

#ifndef IN
#define IN
#define OUT
#endif

typedef enum {

    Num8,
    Num16,
    Flag

} ECU_FIELD_TYPE;


typedef struct _ECU_FIELD_DESCRIPTOR {

    ULONG FieldID;   // 1 based unique id for field

    //
    //  Byte + bit of the init response which indicate that this value is supported
    //  by the current ECU.
    //

    ULONG InitResponseSupportedByte;
    UCHAR InitResponseSupportedBit;

    //
    //  Description of the associated field in the ECU (note that address is implied
    //  by the reference from the current ECU memory layout table).
    //

    ECU_FIELD_TYPE FieldType;
    UCHAR FieldBitMask;     // for flag type fields

    //
    //  How to process the raw data into an output value, and what the units are called.
    //
    //  The subtract / multiply / divide will be performed in that order to produce the
    //  reported
    //

    PUCHAR FieldLongName;
    PUCHAR FieldShortName;
    PUCHAR FieldUnitsString;

    ULONG Subtract;
    double Multiplier;
    double Divisor;

}
ECU_FIELD_DESCRIPTOR, *PECU_FIELD_DESCRIPTOR;

//
//  Structures to describe a particular ECU revision and it's supported fields.
//

#define ECU_MEMORY_MAPPED_SIZE 0x130

typedef struct _ECU_MEM_ADDR_DESC {

    ULONG Addr;
    ULONG FieldID;
}
ECU_MEM_ADDR_DESC, *PECU_MEM_ADDR_DESC;

#define ECU_ID_SIZE 5

typedef struct _ECU_DESCRIPTOR {

    PUCHAR Name;

    //
    //  Array of addresses + corresponding fields supported..
    //

    ECU_MEM_ADDR_DESC MemoryMap[];
}
ECU_DESCRIPTOR, *PECU_DESCRIPTOR;

typedef struct _ECU_ID_MAPPING {

    UCHAR RawId[5];
    PECU_DESCRIPTOR Descriptor;

} ECU_ID_MAPPING, *PECU_ID_MAPPING;


//
//  Static tables of known ECU, supported fields etc.
//

extern ECU_FIELD_DESCRIPTOR EcuFieldTable[];
extern ECU_ID_MAPPING EcuIdMappings[];

//
//  Structures for communicating with the ECU.
//

//
//  Force byte alignment of fields.
//

#pragma pack(1)

#define ECU_OP_INIT         0xBF
#define ECU_OP_READ_MEM     0xA0
#define ECU_OP_READ_ADDR    0xA8
#define ECU_OP_WRITE        0xB0
#define ECU_OP_WRITE_ADDR   0xB8

#define ECU_ADDR            0x10
#define PC_ADDR             0xF0

typedef struct _PACKET {

    UCHAR Signature;  // always 0x80
    UCHAR DestAddr;
    UCHAR SourceAddr;
    UCHAR DataByteCount;
    UCHAR Data[0];
}
PACKET, *PPACKET;

#define MAX_PACKET_SIZE (sizeof(PACKET) + 256 + 1)

typedef struct _PACKET_WBUF {

    union {

        PACKET Pkt;
        UCHAR Raw[MAX_PACKET_SIZE];
    };
}
PACKET_WBUF, *PPACKET_WBUF;


#define CHECKSUM_SIZE 1
#define PACKET_SIZE( PacketDataSize)  (sizeof(PACKET) + (PacketDataSize))
#define PACKET_SIZE_CHK( PacketDataSize) (PACKET_SIZE((PacketDataSize)) + CHECKSUM_SIZE)

//
//  ECU INIT request details.
//

//  Init response packet length EXCLUDING packet header & checksum byte

#define ECU_INIT_RESPONSE_DATA_LENGTH     57

//  Data length excluding padding.

#define ECU_INIT_RESPONSE_INIT_DATA_LEN   51

typedef struct _ECU_INIT_RESPONSE {

    UCHAR Padding;
    UCHAR Unknown[3];
    UCHAR EcuID[5];
    UCHAR Data[ ECU_INIT_RESPONSE_INIT_DATA_LEN];

} ECU_INIT_RESPONSE, *PECU_INIT_RESPONSE;

//
//  ECU READ/WRITE stuff.
//
//  ECU is big-endian and uses 24bit addresses...
//

#define ECU_ADDR_LEN 3

#define ECU_READ_ADDR_REQUEST_SIZE( ADDRS)  (PACKET_SIZE_CHK(((ADDRS)*ECU_ADDR_LEN)+2))
#define ECU_READ_ADDR_RESPONSE_SIZE( ADDRS) (PACKET_SIZE_CHK((ADDRS)+1))

//
//  Set default packing again.
//

#pragma pack()

// TODO: Temporary
extern ECU_DESCRIPTOR ECU_GENERIC_WRX;



//
//	read request size = packet + chk + 3b(addr) + read code + pad + length
//  Read response size = packet + chksum + bytes read + pad byte
//

#define ECU_READ_MEM_RESPONSE_SIZE( BytesRead) PACKET_SIZE_CHK((BytesRead) + 1)
#define ECU_READ_MEM_REQUEST_SIZE PACKET_SIZE_CHK( 6)

#endif

