/*

This program and source is licensed under the

    GNU GENERAL PUBLIC LICENSE

as in the included LICENSE.TXT or located at
http://www.gnu.org/licenses/gpl.txt

*/

#include <stdio.h>
#include <math.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>

#include "ecu.h"

#include "ecucom.h"
#include "trace.h"

#include "ecucore.h"

typedef enum {

    CoreStateNotInitialized,
    CoreStateNotConnected,
    CoreStateConnectedIdle,
    CoreStateCollectingData,
    CoreStateDumpingRom

} ECU_CORE_STATE;


static ECU_CORE_STATE CurrentState = CoreStateNotInitialized;
static PECU_DESCRIPTOR CurrentEcu = NULL;

//
//  Store the ECU init response for later use.
//

static UCHAR EcuInitResponseData[256];


BOOL
ValidateResponse(
    IN PPACKET ResponseBuffer,
    IN ULONG TotalDataLength,
    OUT PPACKET *ResponsePacket
    )
{
    PPACKET Response;

    *ResponsePacket = NULL;

    DumpHex(TRACE_ECU, ResponseBuffer, TotalDataLength);

    //
    //  We expect the buffer to contain an echo of our request packet.
    //

    if ((ResponseBuffer->Signature != 0x80) ||
        (ResponseBuffer->SourceAddr != PC_ADDR ) ||
        (ResponseBuffer->DestAddr != ECU_ADDR )) {

        TracePrint(TRACE_ERROR,"Response echo malformed\n");
        return FALSE;
    }

    Response = (PPACKET)(((PUCHAR)ResponseBuffer) +
                         PACKET_SIZE_CHK( ResponseBuffer->DataByteCount));

    if ((Response->Signature != 0x80) ||
        (Response->SourceAddr != ECU_ADDR ) ||
        (Response->DestAddr != PC_ADDR )) {

        TracePrint(TRACE_ERROR,"Response malformed\n");
        DumpHex(TRACE_ERROR, Response, sizeof(PACKET));
        return FALSE;
    }

    // TODO: Validate checksum?

    *ResponsePacket = Response;
    return TRUE;
}



BOOL
EcuCoreReceivePacket(
    IN PVOID Buffer,
    IN ULONG ReceiveDataLen,
    OUT PPACKET *ResponsePacket
    )
{
    ULONG Error;
    PPACKET LocalResponse;

    *ResponsePacket = NULL;

    if (!EcuComReceive( Buffer, ReceiveDataLen, &Error)) {

        TracePrint(TRACE_ERROR,"Failed to receive from ECU, %d\n", Error);
        return FALSE;
    }

    TracePrint(TRACE_ECU,"Received packet\n");

    if (!ValidateResponse( Buffer, ReceiveDataLen, &LocalResponse)) {

        return FALSE;
    }

    TracePrint(TRACE_ECU,"Response data:\n");
    DumpHex(TRACE_ECU, LocalResponse->Data, LocalResponse->DataByteCount);

    *ResponsePacket = LocalResponse;
    return TRUE;
}



ULONG
EcuCorePreparePacket(
    IN OUT PPACKET_WBUF PacketBuffer
    )
{
    UCHAR Checksum = 0;
    ULONG I;

    //
    //  Copy in data, fill in fixed fields and calculate checksum.
    //

    PacketBuffer->Pkt.Signature = 0x80;
    PacketBuffer->Pkt.DestAddr = ECU_ADDR;
    PacketBuffer->Pkt.SourceAddr = PC_ADDR;

    for (I = 0; I < PACKET_SIZE( PacketBuffer->Pkt.DataByteCount); I++) {

        Checksum += PacketBuffer->Raw[I];
    }

    PacketBuffer->Pkt.Data[ PacketBuffer->Pkt.DataByteCount] = Checksum;

    return (PACKET_SIZE( PacketBuffer->Pkt.DataByteCount) + CHECKSUM_SIZE);
}



BOOL
EcuCoreConnectEcu(
    OUT PECU_DESCRIPTOR *Ecu,
    OUT PUCHAR *InitResponseData
    )
{
    PACKET_WBUF Packet;
    UCHAR ResponseBuffer[256];
    PPACKET ResponsePacket;
    ULONG Error;
    ULONG Retries;
    BOOL Result = FALSE;

    if (CurrentState != CoreStateNotConnected) {

        return FALSE;
    }

    *Ecu = NULL;

    Packet.Pkt.DataByteCount = 1;
    Packet.Pkt.Data[0] = ECU_OP_INIT;

    for (Retries = 1; Retries < 4; Retries++) {

        TracePrint(TRACE_INFO,"Initializing ECU connection (attempt %d)\n", Retries);

        EcuComPurge();

        EcuCorePreparePacket( &Packet);

        if (!EcuComSend( &Packet, PACKET_SIZE_CHK( Packet.Pkt.DataByteCount), &Error)) {

            TracePrint(TRACE_ERROR,"Failed to send ECU init, %d\n", Error);
            return FALSE;
        }

        TracePrint(TRACE_INFO,"Sent INIT, waiting for response.\n");

        //
        //  ECU echos back our packet, so that needs to be included in the receive.
        //

        if (EcuCoreReceivePacket( ResponseBuffer,
                                  PACKET_SIZE_CHK( Packet.Pkt.DataByteCount)
                                    + PACKET_SIZE_CHK( ECU_INIT_RESPONSE_DATA_LENGTH),
                                  &ResponsePacket)) {

            Result = TRUE;
            break;
        }

        //
        //  Wait 500ms and try again.
        //

        Sleep(500);
    }

    if (Result) {

        //
        //  Lookup the ECU ID in the table of known ECUs.  None found, default to
        //  the 'generic' ecu description.
        //

        // TODO:

        *Ecu = CurrentEcu = &ECU_GENERIC_WRX;
        *InitResponseData = EcuInitResponseData;

        memcpy( EcuInitResponseData, ResponsePacket->Data, ResponsePacket->DataByteCount);

        CurrentState = CoreStateConnectedIdle;
    }

    return Result;
}


BOOL
EcuCoreInit(
    IN ULONG ComPort
    )
{
    BOOL Result;

    if (CurrentState != CoreStateNotInitialized) {

        return FALSE;
    }

    Result = EcuComInitialize( ComPort);

    if (Result) {

        CurrentState = CoreStateNotConnected;
    }

    // TODO: Validate ECU tables for no dup fieldids, addresses, etc.

    return Result;
}


VOID
EcuCoreUnInit(
    )
{
    if (CurrentState == CoreStateNotInitialized) {

        return;
    }

    // TODO: Stop active logging, etc.

    EcuComUnInitialize();

    CurrentState = CoreStateNotInitialized;
}


PECU_FIELD_DESCRIPTOR
EcuCoreFieldFromFieldId(
    IN ULONG Id
    )
{
    ULONG Index = 0;
    PECU_FIELD_DESCRIPTOR Field = NULL;

    do {

        if (EcuFieldTable[ Index].FieldID == Id) {

            Field = &EcuFieldTable[ Index];
            break;
        }

        Index += 1;
    }
    while (EcuFieldTable[ Index].FieldID != 0);

    return Field;
}

PUCHAR
EcuCoreUnitsFromFieldId(
    IN ULONG Id
    )
{
    ULONG Index = 0;
    PUCHAR Units = NULL;

    do {

        if (EcuFieldTable[ Index].FieldID == Id) {

            Units = EcuFieldTable[ Index].FieldUnitsString;
            break;
        }

        Index += 1;
    }
    while (EcuFieldTable[ Index].FieldID != 0);

    return Units;
}

ULONG
EcuCoreFieldIdFromFieldShortName(
    IN PECU_DESCRIPTOR Ecu,
    IN UCHAR ShortName[]
    )
//
//  Walk the ECU decriptor looking for a supported field with matching SHORT NAME.  Don't
//
//
{
    ULONG Index = 0;
    ULONG FieldId = ECU_CORE_INVALID_FIELD_ID;

    do {

        if (!_stricmp( EcuCoreFieldFromFieldId( Ecu->MemoryMap[ Index].FieldID)->FieldShortName,
                       ShortName)) {

            FieldId = Ecu->MemoryMap[ Index].FieldID;
            break;
        }

        Index += 1;
    }
    while (Ecu->MemoryMap[ Index].FieldID != 0);

    return FieldId;
}


//
//  Implementation of the data collection engine
//


VOID
AddressToEcuAddress(
    IN ULONG Address,
    OUT PUCHAR Target
    )
{
    Target[0] = (UCHAR)(Address >> 16);
    Target[1] = (UCHAR)(Address >> 8);
    Target[2] = (UCHAR)(Address & 0xFF);
}


#define DATA_BUFFER_SIZE                (64*1024)
#define MAX_OUTSTANDING_ECU_REQUESTS    1



typedef struct _ECU_CORE_COLLECTION_CONTEXT {

    PECU_CORE_SAMPLE_SET SampleSet;
    PECU_CORE_SAMPLE_DESCRIPTOR SampleList;
    ULONG SampleListSize;

    ULONG NextOffsetToWrite;
    ULONG NextOffsetToRead;

    ULONG BufferWrappedOffset;

    ULONG BufferWrapCount;
    ULONG ReaderWrapCount;

    //
    //  Buffer used for receiving data from ECU
    //

    UCHAR DataBuffer[ DATA_BUFFER_SIZE];    // TODO: Stamp receive time in buffer before packet

    //
    //  Query issue thread stores size of issued queries here for receiver thread
    //  to use.  Value is number of memory locations queried.
    //

    ULONG NumValuesInQuery[ MAX_OUTSTANDING_ECU_REQUESTS];

    DWORD StartTickCount;
    ULONG SampleCycle;

    //
    //  Threads and synchronization.
    //

    HANDLE RequestorThread;
    HANDLE ReceiverThread;

    HANDLE UnReceivedPacketsSemaphore;
    HANDLE OutstandingPacketsSemaphore;
    HANDLE ReceivedPacketsSemaphore;

    BOOL Abort;

    // starttime
    //

} ECU_CORE_COLLECTION_CONTEXT, *PECU_CORE_COLLECTION_CONTEXT;


ECU_CORE_COLLECTION_CONTEXT CollectionContext;


DWORD
WINAPI
EcuCoreDataCollectRequestor(
    PVOID Parameter
    )
{
    ULONG SampleCycle = 1;
    PACKET_WBUF QueryPacket;
    ULONG DataLength;
    ULONG AddressesQueried;
    ULONG PacketLength;
    ULONG I;
    ULONG Error;

    QueryPacket.Pkt.Data[0] = ECU_OP_READ_ADDR;
    QueryPacket.Pkt.Data[1] = 0;                // padding

    do {

        WaitForSingleObject( CollectionContext.OutstandingPacketsSemaphore, INFINITE);

        if (CollectionContext.Abort) {

            break;
        }

        DataLength = 2;
        AddressesQueried = 0;

        for (I = 0; I < CollectionContext.SampleListSize; ++I) {

            //
            //  If we're not supposed to sample this field this cycle, don't.
            //

            if (0 != (SampleCycle % CollectionContext.SampleList[ I].Frequency)) {

                continue;
            }

            // TODO: FLAG FIELDS!

            AddressToEcuAddress( CollectionContext.SampleList[ I].EcuAddress,
                                 &(QueryPacket.Pkt.Data[ DataLength]));

            if (CollectionContext.SampleList[ I].Field->FieldType == Num16) {

                AddressesQueried += 1;
                DataLength += ECU_ADDR_LEN;

                AddressToEcuAddress( CollectionContext.SampleList[ I].EcuAddress + 1,
                                     &(QueryPacket.Pkt.Data[ DataLength]));
            }

            AddressesQueried += 1;
            DataLength += ECU_ADDR_LEN;
        }

        //
        //  Remember how many addresses we queried for the receiver thread.
        //

        CollectionContext.NumValuesInQuery[ SampleCycle % MAX_OUTSTANDING_ECU_REQUESTS] = AddressesQueried;

        if (DataLength > 255) {

            TracePrint(TRACE_ERROR,"Internal error - packet data size too large\n");
            exit(1);
        }

        QueryPacket.Pkt.DataByteCount = (UCHAR)DataLength;

        PacketLength = EcuCorePreparePacket( &QueryPacket);

        //
        //  Tell receiver thread it can wait on a packet now.
        //

        ReleaseSemaphore( CollectionContext.UnReceivedPacketsSemaphore, 1, NULL);

        //
        //  Send the packet!
        //

        if (!EcuComSend( &QueryPacket, PacketLength, &Error)) {

            TracePrint(TRACE_ERROR,"DataCollector - failed to send query packet\n");
            exit(1); // TODO: clean up
        }

        SampleCycle += 1;

    }
    while (TRUE);

    return 0;
}


DWORD
WINAPI
EcuCoreDataCollectReceiver(
    PVOID Parameter
    )
{
    ULONG SampleCycle = 1;
    ULONG NumValuesExpected;
    ULONG ExpectedTotalResponseSize;
    PPACKET Response;

    do {

        //
        //  Wait for sender thread to send a packet.
        //

        WaitForSingleObject( CollectionContext.UnReceivedPacketsSemaphore, INFINITE);

        NumValuesExpected = CollectionContext.NumValuesInQuery[ SampleCycle % MAX_OUTSTANDING_ECU_REQUESTS];

        ExpectedTotalResponseSize = ECU_READ_ADDR_REQUEST_SIZE( NumValuesExpected) +
                                    ECU_READ_ADDR_RESPONSE_SIZE( NumValuesExpected);

        if (!EcuCoreReceivePacket( &CollectionContext.DataBuffer[ CollectionContext.NextOffsetToWrite],
                                   ExpectedTotalResponseSize,
                                   &Response)) {

            TracePrint(TRACE_ERROR,"Data collection - Receive packet failed\n");
            exit(1); // TODO: Clean  up error handling
        }

        if (CollectionContext.Abort) {

            break;
        }

        //
        //  Free up the requestor to send another packet.
        //

        if (!ReleaseSemaphore( CollectionContext.OutstandingPacketsSemaphore, 1, NULL)) {

            CollectionContext.Abort = TRUE;
            break;
        }

        //
        //  Update buffer pointer
        //

        CollectionContext.NextOffsetToWrite += ExpectedTotalResponseSize;

        //
        //  Adjust if we hit end of buffer
        //

        if (CollectionContext.NextOffsetToWrite >= (DATA_BUFFER_SIZE - MAX_PACKET_SIZE)) {

            CollectionContext.NextOffsetToWrite = 0;
            CollectionContext.BufferWrapCount += 1;

            TracePrint(TRACE_BUFFER,"Receiver wrap %d\n", CollectionContext.BufferWrapCount);
        }

        if ((CollectionContext.BufferWrapCount > CollectionContext.ReaderWrapCount) &&
            ((CollectionContext.NextOffsetToWrite + MAX_PACKET_SIZE) > CollectionContext.NextOffsetToRead)) {

            TracePrint(TRACE_ERROR,"Receiving data faster than reader processing, buffer full\n");
            CollectionContext.Abort = TRUE;
        }

        //
        //  Free up the client thread to parse the data we just received.
        //

        if (!ReleaseSemaphore( CollectionContext.ReceivedPacketsSemaphore, 1, NULL)) {

            CollectionContext.Abort = TRUE;
            break;
        }

        SampleCycle += 1;
    }
    while (!CollectionContext.Abort);

    return 0;
}


BOOL
EcuCoreCollectData(
    IN     PECU_CORE_SAMPLE_SET SampleSet,
    IN     PECU_CORE_SAMPLE_NOTIFY_CALLBACK SampleNotifyCallBack
    )
{
    ULONG I, J;
    BOOL Match;
    ULONG QSize, RSize;
    PPACKET Response;
    ULONG SampleCycle = 1;
    ULONG ECUResponseValueIndex;
    ULONG BytesFromEcu = 0;
    PECU_FIELD_DESCRIPTOR Field;
    PECU_CORE_SAMPLE_DESCRIPTOR SampleList = SampleSet->SampleList;
    double Value;

    if (CurrentState != CoreStateConnectedIdle) {

        TracePrint(TRACE_ERROR,"ECU connection not initialized or busy\n");
        return FALSE;
    }

    if (SampleSet->NumFieldsToSample > ECU_MAX_SAMPLE_FIELDS) {

        TracePrint(TRACE_ERROR,"Too many sample fields supplied - %d max is %d\n",
                   SampleSet->NumFieldsToSample, ECU_MAX_SAMPLE_FIELDS);
        return FALSE;
    }

    //
    //  Map the requested fields to actual ECU addresses
    //

    for (I = 0; I < SampleSet->NumFieldsToSample; ++I) {

        //
        //  Walk the fields defined for this ECU revision looking for a match.
        //

        J = 0;
        Match = FALSE;

        do {

            if (CurrentEcu->MemoryMap[ J].FieldID == SampleList[ I].FieldId) {

                SampleList[ I].EcuAddress = CurrentEcu->MemoryMap[ J].Addr;
                SampleList[ I].Field = EcuCoreFieldFromFieldId( SampleList[ I].FieldId);
                // TODO: FLAG FIELDs
                BytesFromEcu += (Num8 == SampleList[ I].Field->FieldType) ? 1 : 2;
                Match = TRUE;
                break;
            }

            J++;
        }
        while (CurrentEcu->MemoryMap[ J].FieldID != 0);

        if (!Match) {

            TracePrint(TRACE_ERROR,"Requested ECU field id %d (\'%s\') not mapped for this ECU revision\n",
                       SampleList[ I].FieldId,
                       EcuCoreFieldFromFieldId( SampleList[ I].FieldId)->FieldLongName);
            return FALSE;
        }
    }

    if (0 == BytesFromEcu) {

        TracePrint(TRACE_ERROR,"No ECU sourced fields specified\n");
        return FALSE;
    }
    else if (BytesFromEcu > ECU_MAX_SAMPLE_FIELDS) {

        // TODO: This is wrong - doesn't take into account some samples may be every N cycles...

        TracePrint(TRACE_ERROR,"Too much data requested from ECU\n");
        return FALSE;
    }

    CurrentState = CoreStateCollectingData;

    memset( &CollectionContext, 0, sizeof( CollectionContext));

    CollectionContext.SampleSet = SampleSet;
    CollectionContext.SampleList = SampleList;
    CollectionContext.SampleListSize = SampleSet->NumFieldsToSample;

    // TODO: Cleanup properly in this function if fail at various stages below!!

    CollectionContext.OutstandingPacketsSemaphore = CreateSemaphore( NULL,
                                                                     MAX_OUTSTANDING_ECU_REQUESTS,
                                                                     MAX_OUTSTANDING_ECU_REQUESTS,
                                                                     NULL);
    if (NULL == CollectionContext.OutstandingPacketsSemaphore) {

        TracePrint(TRACE_ERROR,"Failed to create outstanding packets semaphore, error %d\n", GetLastError());
        return FALSE;
    }

    CollectionContext.ReceivedPacketsSemaphore = CreateSemaphore( NULL,
                                                                  0,
                                                                  0xffff,
                                                                  NULL);
    if (NULL == CollectionContext.ReceivedPacketsSemaphore) {

        CloseHandle( CollectionContext.OutstandingPacketsSemaphore);
        TracePrint(TRACE_ERROR,"Failed to create received packets semaphore, error %d\n", GetLastError());
        return FALSE;
    }

    CollectionContext.UnReceivedPacketsSemaphore = CreateSemaphore( NULL,
                                                                    0,
                                                                    0xffff,
                                                                    NULL);
    if (NULL == CollectionContext.UnReceivedPacketsSemaphore) {

        TracePrint(TRACE_ERROR,"Failed to create received packets semaphore, error %d\n", GetLastError());
        return FALSE;
    }

    CollectionContext.ReceiverThread = CreateThread( NULL,
                                                     0,
                                                     EcuCoreDataCollectReceiver,
                                                     NULL,
                                                     0,
                                                     NULL);

    if (NULL == CollectionContext.ReceiverThread) {

        TracePrint(TRACE_ERROR,"Failed to create receiver thread, error %d\n", GetLastError());
        exit(1);
    }

    CollectionContext.RequestorThread = CreateThread( NULL,
                                                      0,
                                                      EcuCoreDataCollectRequestor,
                                                      NULL,
                                                      0,
                                                      NULL);

    if (NULL == CollectionContext.RequestorThread) {

        // TODO: Cleanup properly in tihs function
        TracePrint(TRACE_ERROR,"Failed to create requestor thread, error %d\n", GetLastError());
        exit(1);
    }

    SetThreadPriority( CollectionContext.RequestorThread, THREAD_PRIORITY_ABOVE_NORMAL);
    SetThreadPriority( CollectionContext.ReceiverThread, THREAD_PRIORITY_ABOVE_NORMAL);

    CollectionContext.StartTickCount = GetTickCount();

    do {

        WaitForSingleObject( CollectionContext.ReceivedPacketsSemaphore, INFINITE);

        //
        //  Received data format is [QPACKET][RPACKET]. Skip our query.
        //

        Response = (PVOID)(&CollectionContext.DataBuffer[ CollectionContext.NextOffsetToRead]);

        QSize = PACKET_SIZE_CHK( Response->DataByteCount);

        Response = (PVOID)(&CollectionContext.DataBuffer[ CollectionContext.NextOffsetToRead + QSize]);

        RSize = PACKET_SIZE_CHK( Response->DataByteCount);

        ECUResponseValueIndex = 0;

        //  parse packet data from buffer into caller array

        for (I = 0; I < CollectionContext.SampleListSize; ++I) {

            //
            //  Did we sample this counter this time?
            //

            if (0 == (SampleCycle % CollectionContext.SampleList[ I].Frequency)) {

                //
                //  Grab the data from the ECU response.
                //

                // TODO: Handle FLAG fields

                Field = CollectionContext.SampleList[ I].Field;

                Value = Response->Data[ 1 + ECUResponseValueIndex];

                if (CollectionContext.SampleList[ I].Field->FieldType == Num16) {

                    ECUResponseValueIndex += 1;

                    Value *= 0xff;
                    Value += Response->Data[ 1 + ECUResponseValueIndex];
                }

                //
                //  Transform the raw value into it's final units.
                //

                Value = (Value - Field->Subtract) * Field->Multiplier / Field->Divisor;

                CollectionContext.SampleList[ I].Value = Value;
                CollectionContext.SampleList[ I].UpdatedThisSample = TRUE;
            }
            else {

                // No.

                CollectionContext.SampleList[ I].UpdatedThisSample = FALSE;
                continue;
            }

            ECUResponseValueIndex += 1;
        }

        //
        //  Tell the caller we placed data in their buffer and have them process it.
        //

        // TODO: toime should be from receiver not calculated here
        SampleSet->TimeStamp = GetTickCount() - CollectionContext.StartTickCount;
		SampleSet->SampleNumber = SampleCycle;
        
        CollectionContext.Abort = !(SampleNotifyCallBack)();

        SampleCycle += 1;
        CollectionContext.NextOffsetToRead += QSize + RSize;

        //
        //  Adjust if we hit end of buffer
        //

        if (CollectionContext.NextOffsetToRead >= (DATA_BUFFER_SIZE - MAX_PACKET_SIZE)) {

            CollectionContext.NextOffsetToRead = 0;
            CollectionContext.ReaderWrapCount += 1;

            TracePrint(TRACE_BUFFER,"Reader wrap %d\n", CollectionContext.ReaderWrapCount);
        }
    }
    while (!CollectionContext.Abort);

    WaitForSingleObject( CollectionContext.ReceiverThread, INFINITE);
    WaitForSingleObject( CollectionContext.RequestorThread, INFINITE);

    CloseHandle( CollectionContext.RequestorThread);
    CloseHandle( CollectionContext.ReceiverThread);

    CurrentState = CoreStateConnectedIdle;

    return TRUE;
}


//
//  Functions related to reading/writing memory.
//

BOOL
EcuCoreReadMemory(
    IN ULONG ECUAddress,
    IN UCHAR Length,
    IN OUT PVOID Buffer
    )
{
    PACKET_WBUF Packet;
	UCHAR ReceiveBuffer[ ECU_READ_MEM_REQUEST_SIZE + ECU_READ_MEM_RESPONSE_SIZE( 255)];
    ULONG Size;
    ULONG Error;

    if (CurrentState != CoreStateConnectedIdle) {

        TracePrint(TRACE_ERROR,"ECU connection not initialized or busy\n");
        return FALSE;
    }

	if ((0 == Length) || (Length > 255)){

        TracePrint(TRACE_ERROR,"Attempt to read zero length memory block or > max length\n");
        return FALSE;
    }

    TracePrint(TRACE_INFO,"Reading memory at 0x%x for 0x%x bytes\n",ECUAddress,Length);

    Packet.Pkt.Data[0] = ECU_OP_READ_MEM;
    Packet.Pkt.Data[1] = 0;                // padding

    AddressToEcuAddress( ECUAddress, &Packet.Pkt.Data[2]);

    Packet.Pkt.Data[5] = Length - 1;

    Packet.Pkt.DataByteCount = 6;

    Size = EcuCorePreparePacket( &Packet);

    if (!EcuComSend( &Packet, Size, &Error)) {

        TracePrint(TRACE_ERROR,"Failed to send read memroy request\n");
        return FALSE;
    }

    if (!EcuComReceive( ReceiveBuffer, ECU_READ_MEM_RESPONSE_SIZE( Length), &Error)) {

        TracePrint(TRACE_ERROR,"Failed to receive read memory response\n");
        return FALSE;
    }

	memcpy( Buffer, ReceiveBuffer + ECU_READ_MEM_REQUEST_SIZE + sizeof( PACKET) + 1, Length);

    return TRUE;
}


